##############################################################################
#
# Copyright (c) 2002 Nexedi SARL and Contributors. All Rights Reserved.
#                    Jean-Paul Smets-Solane <jp@nexedi.com>
#                    Sebastien Robin <seb@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from Globals import PersistentMapping
from time import gmtime,strftime # for anchors
from SyncCode import SyncCode

class Signature(SyncCode):
    """
        Holds

        status -- SENT, CONFLICT...

        md5 -- An MD5 value of a given document

        uid -- The UID of the document
        
        id -- the ID of the document

        rid -- the uid of the document on the remote database,
               only needed on the server.

    """

    # Constructor
    def __init__(self,uid):
        self.uid = uid

    def setStatus(self, status):
        """
            set the Status (see SyncCode for numbers)
        """
        self.status = status

    def getStatus(self, status):
        """
            get the Status (see SyncCode for numbers)
        """
        return self.status

    def setRid(self, rid):
        """
            set the rid
        """
        self.rid = rid

    def getRid(self, rid):
        """
            get the rid
        """
        return self.rid

    def setUid(self, uid):
        """
            set the uid
        """
        self.uid = uid

    def getUid(self, uid):
        """
            get the uid
        """
        return self.uid

class Subscription(SyncCode):
    """
        Subscription hold the definition of a master ODB
        from/to which a selection of objects will be synchronised

        Subscription defined by::

        publication_url -- a URI to a publication

        subsribtion_url -- URL of ourselves

        destination_path -- the place where objects are stored

        query      -- a query which defines a local set of documents which
                      are going to be synchronised

        xml_mapping -- a PageTemplate to map documents to XML

        Subscription also holds private data to manage
        the synchronisation. We choose to keep an MD5 value for
        all documents which belong to the synchronisation process::

        signatures -- a dictionnary which contains the signature
                     of documents at the time they were synchronized

        session_id -- it defines the id of the session
                  with the server.

        last_anchor - it defines the id of the last synchronisation

        next_anchor - it defines the id of the current synchronisation

    """

    signatures = PersistentMapping()
    if signatures is None:
        signatures = {}

    # Constructor
    def __init__(self, id, publication_url, subscription_url, destination_path, query, xml_mapping):
        """
            We need to create a dictionnary of
            signatures of documents which belong to the synchronisation
            process
        """
        self.id = id
        self.publication_url = (publication_url)
        self.subscription_url = str(subscription_url)
        self.destination_path = str(destination_path)
        self.query = query
        self.xml_mapping = xml_mapping
        self.anchor = None
        self.session_id = 0
        self.signatures = {}
        self.last_anchor = '00000000T000000Z'
        self.next_anchor = '00000000T000000Z'
        #self.signatures = PersitentMapping()

    # Accessors
    def getRemoteId(self, id, path=None):
        """
            Returns the remote id from a know local id
            Returns None if...
            path allows to implement recursive sync
        """
        pass

    def getLocalId(self, rid, path=None):
        """
            Returns the local id from a know remote id
            Returns None if...
        """
        pass

    def getId(self):
        """
            return the ID
        """
        return self.id

    def setId(self, id):
        """
            set the ID
        """
        self.id = id

    def getQuery(self):
        """
            return the query
        """
        return self.query

    def setQuery(self, query):
        """
            set the query
        """
        self.query = query

    def getPublicationUrl(self):
        """
            return the publication url
        """
        return self.publication_url

    def setPublicationUrl(self, publication_url):
        """
            return the publication url
        """
        self.publication_url = publication_url

    def getXML_Mapping(self):
        """
            return the xml mapping
        """
        return self.xml_mapping

    def setXML_Mapping(self, xml_mapping):
        """
            return the xml mapping
        """
        self.xml_mapping = xml_mapping

    def getSubscriptionUrl(self):
        """
            return the subscription url
        """
        return self.subscription_url

    def setSubscriptionUrl(self, subscription_url):
        """
            set the subscription url
        """
        self.subscription_url = subscription_url

    def getDestinationPath(self):
        """
            return the destination path
        """
        return self.destination_path

    def setDestinationPath(self, destination_path):
        """
            set the destination path
        """
        self.destination_path = destination_path

    def getSubscription(self):
        """
            return the current subscription
        """
        return self

    def getSessionId(self):
        """
            return the session id
        """
        self.session_id += 1
        return self.session_id

    def getLastAnchor(self):
        """
            return the id of the last synchronisation
        """
        return self.last_anchor

    def getNextAnchor(self):
        """
            return the id of the current synchronisation
        """
        return self.next_anchor

    def setLastAnchor(self, last_anchor):
        """
            set the value last anchor
        """
        self.last_anchor = last_anchor

    def setNextAnchor(self, next_anchor):
        """
            set the value next anchor
        """
        # We store the old next anchor as the new last one
        self.last_anchor = self.next_anchor
        self.next_anchor = next_anchor

    def NewAnchor(self):
        """
            set a new anchor
        """
        self.last_anchor = self.next_anchor
        self.next_anchor = strftime("%Y%m%dT%H%M%SZ", gmtime())

    def ResetAnchors(self):
        """
            reset both last and next anchors
        """
        self.last_anchor = self.NULL_ANCHOR
        self.next_anchor = self.NULL_ANCHOR

    def addSignature(self, signature):
        """
            add a Signature to the subscription
        """
        self.signatures[signature.uid] = signature

    def delSignature(self, uid):
        """
            add a Signature to the subscription
        """
        del self.signatures[uid]

    def checkSignature(self, uid):
        """
            Check if there's a signature with this id
        """
        return self.signatures.has_key(uid)

    def ResetAllSignatures(self):
        """
            Reset all signatures
        """
        self.signatures = {}
