# Copyright (c) 2002 Nexedi SARL and Contributors. All Rights Reserved.
#                    Sebastien Robin <seb@nexedi.com>
#                    Jean-Paul Smets-Solane <jp@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

"""\
ERP portal_synchronizations tool.
"""

from OFS.SimpleItem import SimpleItem
from OFS.Folder import Folder
from Products.CMFCore.utils import UniqueObject
from Globals import InitializeClass, DTMLFile, PersistentMapping, Persistent
from AccessControl import ClassSecurityInfo, getSecurityManager
from Products.CMFCore import CMFCorePermissions
from Products.ERP5SyncML import _dtmldir
from Publication import Publication,Subscriber
from Subscription import Subscription,Signature
from xml.dom.ext.reader.Sax2 import FromXmlStream, FromXml
import smtplib # to send emails
from Products.ERP5.Document.Organisation import Organisation
from Products.ERP5.Document.Telephone import Telephone
from XMLSyncUtils import *
#import sys
#import StringIO
import string
import commands
from zLOG import *

from Conduit.ERP5Conduit import ERP5Conduit

class SynchronizationError( Exception ):
    pass

class SynchronizationTool( UniqueObject, SimpleItem ):
    """
        This tool implements the synchronization algorithm
    """


    id              = 'portal_synchronizations'
    meta_type       = 'ERP5 Synchronizations'

    security = ClassSecurityInfo()

    #
    #   Default values.
    #
    list_publications = PersistentMapping()
    if list_publications is None:
        list_publications = []
    list_subscriptions = PersistentMapping()
    if list_subscriptions is None:
        list_subscriptions = []

    # Do we want to use emails ?
    #email = None
    email = 1

    def __init__( self ):
        self.list_publications = []
        self.list_subscriptions = []

    #
    #   ZMI methods
    #
    manage_options = ( ( { 'label'      : 'Overview'
                         , 'action'     : 'manage_overview'
                         }
                       #, { 'label'      : 'Content'
                       #  , 'action'     : 'manage_main'
                       #  }
                       , { 'label'      : 'Publications'
                         , 'action'     : 'managePublications'
                         }
                       , { 'label'      : 'Subscriptions'
                         , 'action'     : 'manageSubscriptions'
                         }
                       #, { 'label'      : 'Properties'
                       #  , 'action'     : 'propertiesForm'
                       #  }
                       )
                     + SimpleItem.manage_options
                     )

    security.declareProtected( CMFCorePermissions.ManagePortal
                             , 'manage_overview' )
    manage_overview = DTMLFile( 'dtml/explainSynchronizationTool', globals() )

    security.declareProtected( CMFCorePermissions.ManagePortal
                             , 'managePublications' )
    managePublications = DTMLFile( 'dtml/managePublications', globals() )

    security.declareProtected( CMFCorePermissions.ManagePortal
                             , 'addPublicationsForm' )
    addPublicationsForm = DTMLFile( 'dtml/addPublications', globals() )

    security.declareProtected( CMFCorePermissions.ManagePortal
                             , 'manageSubsciptions' )
    manageSubscriptions = DTMLFile( 'dtml/manageSubscriptions', globals() )

    security.declareProtected( CMFCorePermissions.ManagePortal
                             , 'addSubscriptionsForm' )
    addSubscriptionsForm = DTMLFile( 'dtml/addSubscriptions', globals() )

    security.declareProtected( CMFCorePermissions.ManagePortal
                             , 'editProperties' )
    def editProperties( self
                      , publisher=None
                      , REQUEST=None
                      ):
        """
            Form handler for "tool-wide" properties (including list of
            metadata elements).
        """
        if publisher is not None:
            self.publisher = publisher

        if REQUEST is not None:
            REQUEST[ 'RESPONSE' ].redirect( self.absolute_url()
                                        + '/propertiesForm'
                                        + '?manage_tabs_message=Tool+updated.'
                                        )

    def addPublications(self, id, publication_url, destination_path, query, xml_mapping, RESPONSE=None):
        """
            create a new publication
        """
        pub = Publication(id, publication_url, destination_path, query, xml_mapping)
        # TODO : see the problem when the list is not initialized
        if len(self.list_publications) == 0:
            self.list_publications = []
        # we can't use append(), otherwise the persistent machinery don't save the change
        self.list_publications = self.list_publications + [pub]
        if RESPONSE is not None:
            RESPONSE.redirect('managePublications')

    def addSubscriptions(self, id, publication_url, subscription_url, destination_path, query, xml_mapping, RESPONSE=None):
        """
            create a new subscription
        """
        sub = Subscription(id, publication_url, subscription_url, destination_path, query, xml_mapping)
        # TODO : see the problem when the list is not initialized
        if len(self.list_subscriptions) == 0:
            self.list_subscriptions = []
        # we can't use append(), otherwise the persistent machinery don't save the change
        self.list_subscriptions = self.list_subscriptions + [sub]
        if RESPONSE is not None:
            RESPONSE.redirect('manageSubscriptions')

    def editPublications(self, id, publication_url, destination_path, query, xml_mapping, RESPONSE=None):
        """
            modify a publication
        """
        pub = Publication(id, publication_url, destination_path, query, xml_mapping)
        for f in range(len(self.list_publications)):
            if self.list_publications[f].id == id:
                #self.list_publications[f] = pub
                self.list_publications = self.list_publications[0:f] + [pub] + self.list_publications[f+1:len(self.list_publications)]
                break
        if RESPONSE is not None:
            RESPONSE.redirect('managePublications')

    def editSubscriptions(self, id, publication_url, subscription_url, destination_path, query, xml_mapping, RESPONSE=None):
        """
            modify a subscription
        """
        sub = Subscription(id, publication_url, subscription_url, destination_path, query, xml_mapping)
        for f in range(len(self.list_subscriptions)):
            if self.list_subscriptions[f].id == id:
                #self.list_subscriptions[f] = pub
                self.list_subscriptions = self.list_subscriptions[0:f] + [sub] + self.list_subscriptions[f+1:len(self.list_subscriptions)]
                break
        if RESPONSE is not None:
            RESPONSE.redirect('manageSubscriptions')

    def deletePublications(self, id, RESPONSE=None):
        """
            delete a publication
        """
        for f in range(len(self.list_publications)):
            if self.list_publications[f].id == id:
                # we can't use pop(), otherwise the persistent machinery don't save the change
                self.list_publications = self.list_publications[0:f] + self.list_publications[f+1:len(self.list_publications)]
                break
        if RESPONSE is not None:
            RESPONSE.redirect('managePublications')

    def deleteSubscriptions(self, id, RESPONSE=None):
        """
            delete a subscription
        """
        for f in range(len(self.list_subscriptions)):
            if self.list_subscriptions[f].id == id:
                # we can't use pop(), otherwise the persistent machinery don't save the change
                self.list_subscriptions = self.list_subscriptions[0:f] + self.list_subscriptions[f+1:len(self.list_subscriptions)]
                break
        if RESPONSE is not None:
            RESPONSE.redirect('manageSubscriptions')

    def ResetPublications(self, id, RESPONSE=None):
        """
            reset a publication
        """
        for f in range(len(self.list_publications)):
            if self.list_publications[f].id == id:
                # we can't use pop(), otherwise the persistent machinery don't save the change
                self.list_publications[f].ResetAllSubscribers()
                break
        if RESPONSE is not None:
            RESPONSE.redirect('managePublications')

    def ResetSubscriptions(self, id, RESPONSE=None):
        """
            reset a subscription
        """
        for f in range(len(self.list_subscriptions)):
            if self.list_subscriptions[f].id == id:
                # we can't use pop(), otherwise the persistent machinery don't save the change
                self.list_subscriptions[f].ResetAllSignatures()
                self.list_subscriptions[f].ResetAnchors()
                break
        if RESPONSE is not None:
            RESPONSE.redirect('manageSubscriptions')

    def getPublications(self):
        """
            Return a list of publications
        """
        return self.list_publications

    def getSubscriptions(self):
        """
            Return a list of publications
        """
        return self.list_subscriptions

    #
    #   Base methods
    #

    def SubSync(self, id, msg=None, RESPONSE=None):
        """
            This is the synchronization method for the client
        """     
        print "********** SubSync **************"
        print "list_publications"
        print self.list_publications
        print self.list_subscriptions
        for f in range(len(self.list_subscriptions)):
            print self.list_subscriptions[f].id
        # search the good subscription
        for f in range(len(self.list_subscriptions)):
            if self.list_subscriptions[f].id == id:
                break

        # first synchronization
        #if self.list_subscriptions[f].next_anchor == self.list_subscriptions[f].NULL_ANCHOR:
        if self.email is None:
            print "we don't use emails"
            file = open('/tmp/sync_server','r')
            if file.readlines() == []:
                self.SubSyncInit(self.list_subscriptions[f])
            else:
                file.seek(0)
                xml_client = FromXmlStream(file)
                self.SubSyncModif(self.list_subscriptions[f],xml_client)
                file.close()
        else:
            print "we use emails"
            if msg==None:
                self.SubSyncInit(self.list_subscriptions[f])
            else: 
                xml_client = FromXml(msg)
                if CheckAlert(xml_client):
                    self.SubSyncModif(self.list_subscriptions[f],xml_client)
                else:
                    self.SubLastSync(self.list_subscriptions[f],xml_client)

        if RESPONSE is not None:
            RESPONSE.redirect('manageSubscriptions')
            
    def SubLastSync(self, id, xml_client=None, RESPONSE=None):
        """
            Now we can read the last responde of the server
        """     
        print "********** SubLastSync **************"
        # search the good subscription
        for f in range(len(self.list_subscriptions)):
            if self.list_subscriptions[f].id == id:
                break

        # Get the destination folder
        list = self.portal_catalog.searchResults(portal_type = "Organisation", meta_type = "ERP5 Folder")
        for folder in list:
            #print "organisation_folder : "; print folder.id
            #print "parent path : "; print string.join(folder.getObject().aq_parent.getPhysicalPath(), '/')
            destination_path_url = string.join(folder.getObject().getPhysicalPath(), '/')
            print "path : "; print  destination_path_url
            if destination_path_url == self.list_subscriptions[f].destination_path:
                #print "I guess I found the good destination"
                destination_path = folder.getObject()

        print "reading message"
        if self.email is None and xml_client is None:
            file = open('/tmp/sync_server','r')
            xml_client = FromXmlStream(file)
        print "end reading message"

        first_node = xml_client.childNodes[1]

        # Get informations from the header
        client_header = first_node.childNodes[1]
        if client_header.nodeName != "SyncHdr":
            print "This is not a SyncML Header"
        alert_code = getAlertCode(xml_client)

        #xml_client = FromXmlStream(file)
        next_action = getNextSyncAction(xml_client, None)
        print "while next action "
        while next_action != None:

            # Add this object to the signatures, so we know that 
            # this object is synchronized with the client
            signature = Signature(getActionId(next_action))
            signature.setStatus(signature.SYNCHRONIZED)
            print "signatures : "; print self.list_subscriptions[f].signatures
            #self.list_subscriptions[f].addSignature(signature)

            print getActionId(next_action)
            # Search if the object is already in the local ZODB
            print "id en cours : " + getActionId(next_action)
            list = self.portal_catalog.searchResults(id = getActionId(next_action))
            print list
            if len(list) != 0:
                pass # the object already exist
            else:
                if getActionType(next_action) == "ERP5 Organisation":
                    o  = Organisation(getActionId(next_action))
                if getActionType(next_action) == "ERP5 Telephone":
                    o  = Telephone(getActionId(next_action), uid=self.portal_relations.NewUID())
                else:
                    o = None
                    print "Unknow type in SubLastSync"
                if o is not None:
                    print "pourtant on l'ajoute bien"
                    id = getActionId(next_action)
                    destination_path._setObject(id,o)

                # FIXME: We should create object with a generic way, but some errors occurs
                #destination_path.invokeFactory(type_name = getActionType(next_action), id = self.portal_relations.NewUID() )
                #destination_path.invokeFactory(type_name = "Telephone", id = self.portal_relations.NewUID() )


            # applied to the remote object
            if getActionType(next_action) == "ERP5 Organisation":
                o  = Organisation("temp")
            if getActionType(next_action) == "ERP5 Telephone": 
                o  = Telephone("temp", uid=self.portal_relations.NewUID())
            if o is not None:
                conduit = ERP5Conduit()
                conduit.newObject(o,next_action)
                destination_path._setObject("temp",o)
                body = getattr(destination_path.temp,self.list_subscriptions[f].xml_mapping)
                file_temp = open('/tmp/sync_remote_object','w')
                file_temp.write("<tmp>\n")
                file_temp.write(body())
                file_temp.write("</tmp>\n")
                file_temp.close()
                destination_path.manage_delObjects('temp')
                    
            list = self.portal_catalog.searchResults(id = getActionId(next_action))
            # Generate an xml file corresponding to the default_Export
            # applied to the local object
            for ci in list:
                if ci.id == getActionId(next_action):
                    object = ci.getObject()
                    body2 = getattr(object,self.list_subscriptions[f].xml_mapping)
                    file_temp = open('/tmp/sync_local_object','w')
                    file_temp.write("<tmp>\n")
                    file_temp.write(body2())
                    file_temp.write("</tmp>\n")
                    file_temp.close()
        
            print "This is a point of Control 1"
            # Generate the xupdate file
            xupdate = commands.getoutput('xmldiff -zx /tmp/sync_local_object /tmp/sync_remote_object')
            file_temp = open('/tmp/sync_xupdate_client','w')
            file_temp.write(xupdate)
            file_temp.close()
            print "This is a point of Control 2"

            # Then update the object
            file_temp = open('/tmp/sync_xupdate_client','r')
            xupdate_stream =  FromXmlStream(file_temp)
            #print "that's fine"
            conduit.ApplyUpdate(object, xupdate_stream)
            file_temp.close()

            print "et bien alors c'est ici"







            next_action = getNextSyncAction(xml_client, next_action)
        if self.email is None:
            file.close()

        if RESPONSE is not None:
            RESPONSE.redirect('manageSubscriptions')
            
    def PubSync(self, id, msg=None, RESPONSE=None):
        """
            This is the synchronization method for the server
        """     
        print "********** PubSync **************"
        print "list_publications"
        print self.list_publications
        print self.list_subscriptions
        for f in range(len(self.list_publications)):
            print self.list_publications[f].id
            print "list_subscribers"
            print self.list_publications[f].list_subscribers
            #for f2 in self.list_publications[f].list_subscribers:
            #    print f2.id
        # search the good publication
        for f in range(len(self.list_publications)):
            if self.list_publications[f].id == id:
                break

        # Read the request from the client
        if self.email is None:
            print "we don't use emails"
            file = open('/tmp/sync_client','r')
            xml_client = FromXmlStream(file)
            file.close()
        else:
            print "we use emails"
            xml_client = FromXml(msg)

        first_node = xml_client.childNodes[1]

        if first_node.nodeName != "SyncML":
            print "This is not a SyncML message"

        # Get informations from the header
        client_header = first_node.childNodes[1]
        if client_header.nodeName != "SyncHdr":
            print "This is not a SyncML Header"
        for subnode in client_header.childNodes:
            if subnode.nodeType == subnode.ELEMENT_NODE and subnode.nodeName == "Source":
                subscription_url = str(subnode.childNodes[0].data)
        for k in self.list_publications[f].list_subscribers:
            print "another subscriber"
            print k.subscription_url
        # Get the subscriber or create it if not already in the list
        subscriber = self.list_publications[f].getSubscriber(subscription_url) 
        #file.close()
        if subscriber == None:
            print "subscriber is None"
            subscriber = Subscriber(subscription_url)
            print "subscription_url" ; print subscription_url
            # FIXME: Why can't we use the method addSubscriber ??
            #self.list_publications[f].addSubscriber(subscriber)
            self.list_publications[f].list_subscribers = self.list_publications[f].list_subscribers + [subscriber]
            # first synchronization
            self.PubSyncInit(self.list_publications[f],xml_client)
        #else: 
        #    self.PubSyncModif(self.list_publications[f], msg)
        elif CheckAlert(xml_client) is not None:
            print "voici le resultat de check"
            self.PubSyncInit(self.list_publications[f], xml_client)
        else:
            print "voici le resultat autre que check"
            self.PubSyncModif(self.list_publications[f], xml_client)
            

        if RESPONSE is not None:
            RESPONSE.redirect('managePublications')
            
    def SubSyncInit(self, subscription):
        """
            Send the first XML message from the client
        """
        print "************** SubSyncInit ***************"
        cmd_id = 1 # specifies a SyncML message-unique command identifier
        xml = ""
        xml += '<SyncML>\n'

        # syncml header
        xml += SyncMLHeader(subscription.getSessionId(), "1", 
                subscription.getPublicationUrl(), subscription.getSubscriptionUrl())

        # syncml body    
        xml += '  <SyncBody>\n'
        subscription.NewAnchor()
        # alert message
        xml += SyncMLAlert(cmd_id, subscription.SLOW_SYNC, subscription.getPublicationUrl(),
                subscription.getDestinationPath(), subscription.getLastAnchor(), subscription.getNextAnchor())
        cmd_id += 1

        xml += '    <Put>\n'
        xml += '      <CmdID>%s</CmdID>\n' % cmd_id ; cmd_id += 1
        # TODO add the DTD and the value devinf11
        # TODO here should be the client specification
        xml += '    </Put>\n'
        xml += '  </SyncBody>\n'

        xml += '</SyncML>\n'

        if self.email is None:
            print "we don't use email"
            file = open('/tmp/sync_client','w')
            file.write(xml)
            file.close()
        else:
            print "we use email"
            SubSendMail(subscription.subscription_url, subscription.publication_url, xml)
        # TODO Send this file to the server's publication instead of just write to a local file

    def PubSyncInit(self, publication, xml_client):
        """
            Read the client xml message
            Send the first XML message from the server
        """
        print "*********** PubSyncInit **************"

        first_node = xml_client.childNodes[1]

        if first_node.nodeName != "SyncML":
            print "This is not a SyncML message"

        # Get informations from the header
        client_header = first_node.childNodes[1]
        if client_header.nodeName != "SyncHdr":
            print "This is not a SyncML Header"

        for subnode in client_header.childNodes:
            if subnode.nodeType == subnode.ELEMENT_NODE and subnode.nodeName == "Source":
                subscription_url = str(subnode.childNodes[0].data)
        subscriber = publication.getSubscriber(subscription_url) # Get the subscriber or create it if not already in the list
        print "subscriber url" ; print subscriber.subscription_url
            

        # Get informations from the body
        last_anchor = getAlertLastAnchor(xml_client)
        next_anchor = getAlertNextAnchor(xml_client)
        alert = CheckAlert(xml_client)
        alert_code = getAlertCode(xml_client)

        # Check if the last time synchronization is the same as the client one
        if subscriber.getNextAnchor() != last_anchor:
            if last_anchor == None:
                print "anchor null iin PubSyncInit"
            else:
                print "bad anchors in PubSyncInit! " + subscriber.getNextAnchor() + " and " + last_anchor
        else:
            subscriber.setNextAnchor(next_anchor)
        #file.close()

        if alert is not None:
            print "alert is not None"
            # Prepare the xml message for the Sync initialization package
            #file = open('/tmp/sync_init_server','w')
            cmd_id = 1 # specifies a SyncML message-unique command identifier
            xml = ""
            xml += '<SyncML>\n'

            # syncml header
            xml += SyncMLHeader(subscriber.getSessionId(), "1", 
                    subscriber.getSubscriptionUrl(), publication.getPublicationUrl())

            # syncml body    
            xml += '  <SyncBody>\n'
            # alert message
            xml += SyncMLAlert(cmd_id, subscriber.SLOW_SYNC, subscriber.getSubscriptionUrl(),
                        publication.getPublicationUrl(), subscriber.getLastAnchor(), subscriber.getNextAnchor())
            cmd_id += 1
            xml += '  </SyncBody>\n'

            xml += '</SyncML>\n'
            #file.write(xml)
            #file.close()

        else:
            "I'm going to pass in PubSyncInit"
            pass

        print "subscriber url2" ; print subscriber.subscription_url
        if self.email is None:
            print "we don't use email"
            file = open('/tmp/sync_server','w')
            file.write(xml)
            file.close()
        else:
            print "we use email"
            SubSendMail(publication.publication_url, subscriber.subscription_url, xml)

    def SubSyncModif(self, subscription, xml_client):
        """
            Send the client modification, this happens after the Synchronization
            initialization
        """
        print "********** SubSyncModif **************"

        # Get informations from the body
        alert_code = getAlertCode(xml_client)
        print "alert code : " ; print alert_code
        last_anchor = getAlertLastAnchor(xml_client)
        if last_anchor != subscription.getLastAnchor():
            print "bad anchors in SubSyndModif!"
        next_anchor = getAlertNextAnchor(xml_client)
        #subscription.setNextAnchor(next_anchor) # Not necesarry, already that value
        #file.close()

        #file = open('/tmp/sync_modif_client','w')
        cmd_id = 1 # specifies a SyncML message-unique command identifier
        xml = ""
        xml += '<SyncML>\n'

        # syncml header
        xml += SyncMLHeader(subscription.getSessionId(), "1", 
                subscription.getPublicationUrl(), subscription.getSubscriptionUrl())

        # syncml body    
        xml += '  <SyncBody>\n'

        # For a slow sync (first synchronization), we have to send everything
        if alert_code == subscription.SLOW_SYNC:
            print "start slow sync"
            subscription.NewAnchor()
            xml += SyncMLStatus(cmd_id, subscription.getPublicationUrl(), subscription.getDestinationPath(),
                        subscription.SLOW_SYNC, subscription.getNextAnchor())
            cmd_id += 1
            xml += '    <Sync>\n'
            list = self.portal_catalog.searchResults(meta_type = subscription.query)
            for ci in list:
                #try:
                print "pourtant ils sont dans la liste"
                print ci.meta_type
                object = ci.getObject()
                if object is not None:
                    print "cette fois c bon"
                    xml += addXMLObject(cmd_id, object, subscription.xml_mapping)
                    cmd_id += 1
                    # Add this object to the signatures, so we know that 
                    # this object is sent to the server
                    signature = Signature(object.uid)
                    signature.setStatus(signature.SENT)
                    #subscription.addSignature(signature)
                #except:
                #    pass
            xml += '    </Sync>\n'

        # For a two way sync (most of the time), we have to send only objects changed
        if alert_code == subscription.TWO_WAY:
            pass

        xml += '  </SyncBody>\n'
        xml += '</SyncML>\n'
        #file.write(xml)
        #file.close()

        if self.email is None:
            print "we don't use email"
            file = open ('/tmp/sync_client','w')
            file.write(xml)
            file.close()
        else:
            print "we use email"
            SubSendMail(subscription.subscription_url, subscription.publication_url, xml)

    def PubSyncModif(self, publication, xml_client):
        """
            Send the server modification, this happens after the Synchronization
            initialization
        """
        print "******* PubSyncModif ************"
        # Get the destination folder
        list = self.portal_catalog.searchResults(portal_type = "Organisation", meta_type = "ERP5 Folder")
        for folder in list:
            #print "organisation_folder : "; print folder.id
            #print "parent path : "; print string.join(folder.getObject().aq_parent.getPhysicalPath(), '/')
            destination_path_url = string.join(folder.getObject().getPhysicalPath(), '/')
            print "path : "; print  destination_path_url
            if destination_path_url == publication.destination_path:
                #print "I guess I found the good destination"
                destination_path = folder.getObject()

        #xml_client = FromXmlStream(file)
        print "read xml with FromXml"
        #xml_client = FromXml(msg)
        first_node = xml_client.childNodes[1]

        # Get informations from the header
        client_header = first_node.childNodes[1]
        if client_header.nodeName != "SyncHdr":
            print "This is not a SyncML Header"
        
        # get the subscriber
        for subnode in client_header.childNodes:
            if subnode.nodeType == subnode.ELEMENT_NODE and subnode.nodeName == "Source":
                subscription_url = subnode.childNodes[0].data
        subscriber = publication.getSubscriber(subscription_url) # Get the subscriber or create it if not already in the list

        alert_code = getAlertCode(xml_client)

        #xml_client = FromXmlStream(file)
        print "before while"
        next_action = getNextSyncAction(xml_client, None)
        while next_action != None:
            print getActionId(next_action)

            # Add this object to the signatures, so we know that 
            # this object is synchronized with the client
            signature = Signature(getActionId(next_action))
            signature.setStatus(signature.SYNCHRONIZED)
            print "add a signature"
            print subscriber.signatures
            subscriber.addSignature(signature)

            # Search if the object is already in the local ZODB
            print "id en cours : " + getActionId(next_action)
            list = self.portal_catalog.searchResults(id = getActionId(next_action))
            print list
            if len(list) != 0:
                pass # the object already exist
            else:
                if getActionType(next_action) == "ERP5 Organisation":
                    o  = Organisation(getActionId(next_action))
                    print "Organisation"
                if getActionType(next_action) == "ERP5 Telephone":
                    o  = Telephone(getActionId(next_action), uid=self.portal_relations.NewUID())
                    print "Telephone"
                else:
                    o = None
                    print "Unknow type in PubSyncModif"
                if o is not None:
                    id = getActionId(next_action)
                    destination_path._setObject(id,o)
                    
            # Generate an xml file corresponding to the default_Export
            # applied to the remote object
            if getActionType(next_action) == "ERP5 Organisation":
                o  = Organisation("temp")
                print "Organisation"
            if getActionType(next_action) == "ERP5 Telephone":
                o  = Telephone("temp", uid=self.portal_relations.NewUID())
                print "Telephone"
            if o is not None:
                conduit = ERP5Conduit()
                conduit.newObject(o,next_action)
                destination_path._setObject("temp",o)
                body = getattr(destination_path.temp,publication.xml_mapping)
                file_temp = open('/tmp/sync_remote_object','w')
                file_temp.write("<tmp>\n")
                file_temp.write(body())
                file_temp.write("</tmp>\n")
                file_temp.close()
                destination_path.manage_delObjects('temp')

            list = self.portal_catalog.searchResults(id = getActionId(next_action))
            # Generate an xml file corresponding to the default_Export
            # applied to the local object
            for ci in list:
                if ci.id == getActionId(next_action):
                    object = ci.getObject()
                    body2 = getattr(object,publication.xml_mapping)
                    file_temp = open('/tmp/sync_local_object','w')
                    file_temp.write("<tmp>\n")
                    file_temp.write(body2())
                    file_temp.write("</tmp>\n")
                    file_temp.close()

            print "This is a point of Control 1"
            # Generate the xupdate file
            xupdate = commands.getoutput('xmldiff -zx /tmp/sync_local_object /tmp/sync_remote_object')
            file_temp = open('/tmp/sync_xupdate','w')
            file_temp.write(xupdate)
            file_temp.close()
            print "This is a point of Control 2"

            # Then update the object
            file_temp = open('/tmp/sync_xupdate','r')
            xupdate_stream =  FromXmlStream(file_temp)
            #print "that's fine"
            conduit.ApplyUpdate(object, xupdate_stream)
            file_temp.close()

            print "et bien alors c'est ici"

            next_action = getNextSyncAction(xml_client, next_action)

        cmd_id = 1 # specifies a SyncML message-unique command identifier
        xml = ""
        xml += '<SyncML>\n'

        # syncml header
        xml += SyncMLHeader(subscriber.getSessionId(), "1", 
                subscriber.getSubscriptionUrl(), publication.getPublicationUrl())

        # syncml body    
        xml += '  <SyncBody>\n'
        # For a slow sync, we have to send everything
        #if alert_code == subscriber.SLOW_SYNC:
        if 1:
            subscriber.NewAnchor()
            xml += SyncMLStatus(cmd_id, subscriber.getSubscriptionUrl(), "organisation",
                        subscriber.SLOW_SYNC, subscriber.getNextAnchor())
            cmd_id += 1
            xml += '    <Sync>\n'
            list = self.portal_catalog.searchResults(meta_type = publication.query)
            for ci in list:
                #try:
                print "pourtant ils sont dans la liste"
                print ci.meta_type
                object = ci.getObject()
                if object is not None:
                    print "cette fois c bon"
                    if subscriber.checkSignature(object.id) == 0:
                        xml += addXMLObject(cmd_id, object, publication.xml_mapping)
                        cmd_id += 1
                #except:
                #    pass
            xml += '    </Sync>\n'
        xml += '  </SyncBody>\n'

        xml += '</SyncML>\n'
        if self.email is None:
            print "we don't use email"
            file = open('/tmp/sync_server','w')
            file.write(xml)
            file.close()
        else:
            print "we use email"
            SubSendMail(publication.publication_url, subscriber.subscription_url, xml)

InitializeClass( SynchronizationTool )
